@@ -1,13 +1,9 @@
-# encoding: UTF-8
-
 require 'spec_helper'
 
 RSpec.describe Mysql2::Error do
-  let(:client) { Mysql2::Client.new(DatabaseCredentials['root']) }
-
   let(:error) do
     begin
-      client.query("HAHAHA")
+      @client.query("HAHAHA")
     rescue Mysql2::Error => e
       error = e
     end
@@ -28,24 +24,26 @@ RSpec.describe Mysql2::Error do
     let(:valid_utf8) { '造字' }
     let(:error) do
       begin
-        client.query(valid_utf8)
+        @client.query(valid_utf8)
       rescue Mysql2::Error => e
         e
       end
     end
 
-    let(:invalid_utf8) { "\xE5\xC6\x7D\x1F" }
+    let(:invalid_utf8) { ["e5c67d1f"].pack('H*').force_encoding(Encoding::UTF_8) }
     let(:bad_err) do
       begin
-        client.query(invalid_utf8)
+        @client.query(invalid_utf8)
       rescue Mysql2::Error => e
         e
       end
     end
 
+    let(:server_info) do
+      @client.server_info
+    end
+
     before do
-      pending('String#encoding is not defined') unless String.public_method_defined?(:encoding)
-
       # sanity check
       expect(valid_utf8.encoding).to eql(Encoding::UTF_8)
       expect(valid_utf8).to be_valid_encoding
@@ -62,7 +60,15 @@ RSpec.describe Mysql2::Error do
         expect(bad_err.message.encoding).to eql(Encoding::UTF_8)
         expect(bad_err.message).to be_valid_encoding
 
-        expect(bad_err.message).to include("??}\u001F")
+        # MariaDB 10.5 returns a little different error message unlike MySQL
+        # and other old MariaDBs.
+        # https://jira.mariadb.org/browse/MDEV-25400
+        err_str = if server_info[:version].match(/MariaDB/) && server_info[:id] >= 100500
+          "??}\\001F"
+        else
+          "??}\u001F"
+        end
+        expect(bad_err.message).to include(err_str)
       end
     end
 
