@@ -2,11 +2,27 @@ require 'rspec'
 require 'mysql2'
 require 'timeout'
 require 'yaml'
+require 'fiber'
+
 DatabaseCredentials = YAML.load_file('spec/configuration.yml')
 
+if GC.respond_to?(:verify_compaction_references)
+  # This method was added in Ruby 3.0.0. Calling it this way asks the GC to
+  # move objects around, helping to find object movement bugs.
+  if RUBY_VERSION >= "3.2"
+    GC.verify_compaction_references(expand_heap: true, toward: :empty)
+  else
+    GC.verify_compaction_references(double_heap: true, toward: :empty)
+  end
+end
+
 RSpec.configure do |config|
   config.disable_monkey_patching!
 
+  config.expect_with :rspec do |expectations|
+    expectations.max_formatted_output_length = 1200
+  end
+
   def with_internal_encoding(encoding)
     old_enc = Encoding.default_internal
     old_verbose = $VERBOSE
@@ -26,6 +42,7 @@ RSpec.configure do |config|
     @clients ||= []
     @clients << client
     return client unless block_given?
+
     begin
       yield client
     ensure
@@ -36,7 +53,7 @@ RSpec.configure do |config|
 
   def num_classes
     # rubocop:disable Lint/UnifiedInteger
-    0.class == Integer ? [Integer] : [Fixnum, Bignum]
+    0.instance_of?(Integer) ? [Integer] : [Fixnum, Bignum]
     # rubocop:enable Lint/UnifiedInteger
   end
 
@@ -51,15 +68,50 @@ RSpec.configure do |config|
     end
   end
 
-  config.before :each do
-    @client = new_client
+  # A directory where SSL certificates pem files exist.
+  def ssl_cert_dir
+    return @ssl_cert_dir if @ssl_cert_dir
+
+    dir = ENV['TEST_RUBY_MYSQL2_SSL_CERT_DIR']
+    @ssl_cert_dir = if dir && !dir.empty?
+      dir
+    else
+      '/etc/mysql'
+    end
+    @ssl_cert_dir
   end
 
-  config.after :each do
-    @clients.each(&:close)
+  # A host used to create the certificates pem files.
+  def ssl_cert_host
+    return @ssl_cert_host if @ssl_cert_host
+
+    host = ENV['TEST_RUBY_MYSQL2_SSL_CERT_HOST']
+    @ssl_cert_host = if host && !host.empty?
+      host
+    else
+      'mysql2gem.example.com'
+    end
+    @ssl_cert_host
   end
 
-  config.before(:all) do
+  config.before(:suite) do
+    begin
+      new_client
+    rescue Mysql2::Error => e
+      username = DatabaseCredentials['root']['username']
+      database = DatabaseCredentials['root']['database']
+      message = %(
+An error occurred while connecting to the testing database server.
+Make sure that the database server is running.
+Make sure that `mysql -u #{username} [options] #{database}` succeeds by the root user config in spec/configuration.yml.
+Make sure that the testing database '#{database}' exists. If it does not exist, create it.
+)
+      warn message
+      raise e
+    end
+  end
+
+  config.before(:context) do
     new_client do |client|
       client.query %[
         CREATE TABLE IF NOT EXISTS mysql2_test (
@@ -120,4 +172,12 @@ RSpec.configure do |config|
       ]
     end
   end
+
+  config.before(:example) do
+    @client = new_client
+  end
+
+  config.after(:example) do
+    @clients.each(&:close)
+  end
 end
